import json
import pprint
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

################################################################################
class network_info:
    """ Basic information about an FDSN network """

    def __init__(self, info):
        """ Initialize using obs-info network.yaml "network_info" field"""
        self.code = info["code"]
        self.start_date = UTCDateTime(info["start_date"])
        self.end_date = UTCDateTime(info["end_date"])
        self.description = info["description"]
        self.comments = info["comments"] if "comments" in info else []


################################################################################
class equipment_type:
    """ Duplicates StationXML EquipmentType """

    def __init__(self, equipment_dict, debug=False):
        """ Initialize from YAML OBS-info equipment dictionary"""
        self.type = None
        self.description = None
        self.manufacturer = None
        self.model = None
        self.serial_number = None
        self.vendor = None
        self.installation_date = None
        self.removal_date = None
        self.calibration_date = None
        if debug:
            print(equipment_dict)
        for key in equipment_dict:
            if not hasattr(self, key):
                raise NameError('No attribute "{}" in FDSN_EquipmentType'.key)
            else:
                setattr(self, key, equipment_dict[key])

    def merge(self, new):
        """ Merge two EquipmentType objects
    
            Takes values from "new" where they exist
            new should be a 
        """
        if not type(new) == equipment_type:
            print("Tried to merge with a non FDSN_EquipmentType")
            return
        for key in vars(new):
            if getattr(new, key):
                setattr(self, key, getattr(new, key))



""" 
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
        
Should modify the code to treat $ref as a placeholder for the associated object
"""
# Standard library modules
import math as m
import json
import pprint
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

from .instrumentation import instrumentation as oi_instrumentation
from ..instrument_components import instrument_components as oi_instrument_components
from ..misc import FDSN

################################################################################
class instrument:
    """ One instrument from instrumentation.yaml file"""

    def __init__(self, filename, station_instrument, referring_file=None, debug=False):
        """ Load an instrument 
    
        Inputs:
            station_instrument: is an OBS_Station.instrument dictionary
                    station_instrument['reference_code'] must correspond to
                        a key in instrumentation['instruments'])        
        """

        instrumentation = oi_instrumentation(filename, referring_file)
        if debug:
            print(60 * "=")
            print(type(station_instrument))
            pprint.pprint(station_instrument)

        generic = self.__get_generic_instrument(
            instrumentation, station_instrument["reference_code"]
        )

        # SET ATTRIBUTES
        self.basepath = instrumentation.basepath
        self.format_version = instrumentation.format_version
        self.revision = instrumentation.revision
        self.facility = instrumentation.facility
        self.components_file = instrumentation.components_file

        self.reference_code = station_instrument["reference_code"]
        self.serial_number = (
            station_instrument["serial_number"]
            if "serial_number" in station_instrument
            else None
        )
        self.equipment = FDSN.equipment_type(generic["equipment"])
        self.das_components = generic["das_components"]

        # SET SPECIFIC ATTRIBUTES (IF ANY)
        specific = self.__get_specific_instrument(instrumentation)
        if specific:
            self.__load_specific_instrument(specific)
        self.equipment.serial_number = self.serial_number
        self.__update_das_components(station_instrument)

    def __repr__(self):
        return "<  {}: reference_code={}, serial_number={}, {:d} channels >".format(
            __name__, self.reference_code, self.serial_number, len(self.das_components)
        )

    def __get_generic_instrument(self, instrumentation, reference_code):
        generics = instrumentation.instruments["generic"]
        if reference_code not in generics:
            raise NameError(
                '"{}" not in generic instrumentation list'.format(reference_code)
            )
        return generics[reference_code]

    def __get_specific_instrument(self, instrumentation):

        specifics = instrumentation.instruments.get("specific", None)
        if not specifics:
            return None

        if self.reference_code not in specifics:
            return None
        if self.serial_number not in specifics[self.reference_code]:
            return None
        return specifics[self.reference_code][self.serial_number]

    def __load_specific_instrument(self, specific, debug=False):
        reference_code = self.reference_code
        serial_number = self.serial_number
        if "orientation_codes" in specific:
            for or_code, inst_spec in specific["orientation_codes"].items():
                das_comp = self.__find_dc_key(or_code)
                if debug:
                    print(
                        "or_code=",
                        or_code,
                        "das_comp=",
                        das_comp,
                        "inst_spec=",
                        inst_spec,
                    )
                self.__inject_measurement_instrument_parameters(das_comp, inst_spec)
        elif "das_components" in specific:
            for das_comp, inst_spec in specific["das_components"].items():
                if debug:
                    print("das_comp=", das_comp, "inst_spec=", inst_spec)
                self.__inject_measurement_instrument_parameters(das_comp, inst_spec)
        else:
            raise NameError(
                'Neither "orientation_codes" nor "das_components" \
                    found for specific instrument {} {}'.format(
                    reference_code, serial_number
                )
            )

    def __update_das_components(self, station_instrument, debug=False):
        # INCORPORATE SPECIFIC CHANNEL VALUES
        for loc_key, value in station_instrument["channel_codes_locations"].items():
            das_component = value.get("das_component", None)
            dc_key = self.__find_dc_key(loc_key[2], das_component)
            self.__insert_codes(dc_key, loc_key)
            self.__update_das_component(dc_key, value)

    def __find_dc_key(self, orientation_code, das_component=None, debug=False):
        """ finds the das_component corresponding to the orientation code
        
            Also validates that the orientation code is possible and unique, using
            das_component if necessary
        """
        if das_component:
            return das_component
        dc_key = None
        das_orientation_codes = []
        for key, value in self.das_components.items():
            das_orientation_codes.append(value["orientation_code"])
            if value["orientation_code"] == orientation_code:
                if dc_key:
                    raise NameError(
                        '"{}" is a non-unique orientation code '
                        "for this instrument\n"
                        "You must specify das_component"
                        'in each "channel_codes_locations" declaration'
                        "".format(orientation_code)
                    )
                if das_component:
                    if das_component == key:
                        dc_key = key
                else:
                    dc_key = key
        if not dc_key:
            raise NameError(
                "instrument {} : No das_component with orientation code "
                '"{}" found, only {} found'.format(
                    self.reference_code, orientation_code, das_orientation_codes
                )
            )
        return dc_key

    def __insert_codes(self, dc_key, chan_loc, debug=False):
        """ inserts location and seed codes into the das_component
        """
        if debug:
            print(chan_loc)
            print(dc_key)
            print(self.das_components[dc_key])
        self.das_components[dc_key]["band_code"] = chan_loc[0]
        self.das_components[dc_key]["inst_code"] = chan_loc[1]
        self.das_components[dc_key]["location_code"] = chan_loc[4:6]

    def __update_das_component(self, dc_key, chan_values, debug=False):
        """ update a das_component with network:station:channel values """
        dc = self.das_components[dc_key]
        if "sample_rate" in chan_values:
            dc["sample_rate"] = chan_values["sample_rate"]
        if "serial_number" in chan_values:
            dc["equipment"].serial_number = chan_values["serial_number"]
        if "start_date" in chan_values:
            dc["start_date"] = chan_values["start_date"]
        if "end_date" in chan_values:
            dc["end_date"] = chan_values["end_date"]

        for block_type in ["sensor", "datalogger", "preamplifier"]:
            if block_type in chan_values:
                for key in chan_values[block_type]:
                    dc[block_type][key] = chan_values[block_type][key]
        if "datalogger_config" in chan_values:
            dl_config = chan_values["datalogger_config"]
            if debug:
                print("ADDING datalogger_config ({})".format(dl_config))
            dc["datalogger"]["reference_code"] = (
                dc["datalogger"]["reference_code"] + "_" + dl_config
            )

    def __inject_measurement_instrument_parameters(
        self, das_component, instrument_spec, debug=False
    ):
        """ reads and injects specific values into a measurement instrument """
        if debug:
            print(instrument_spec)
        for key, values in instrument_spec.items():
            if "reference_code" in values:
                self.das_components[das_component][key]["reference_code"] = values[
                    "reference_code"
                ]
            if "serial_number" in values:
                self.das_components[das_component][key]["serial_number"] = values[
                    "serial_number"
                ]
            if debug:
                print(chan_loc_code, component)
                pprint.pprint(self.channels[chan_loc_code])

    def load_components(self, components_file, referring_file=None):
        """
        Load components into instrument
        
        components_file = name of components file
        referring_file = file that referred to the components file
                         (for resolving paths)
        """
        components = oi_instrument_components(components_file, referring_file)

        for key in self.das_components:
            self.fill_channel(key, components)
        self.resource_id = self.facility["reference_name"] + components.revision["date"]

    def fill_channel(self, channel_key, components, debug=False):
        """ Replace channel component strings with the actual components 
        
            components is an OBS_Instrument_Components object 
                    (direct from *_instrument_components.yaml)
        """
        channel = self.das_components[channel_key]
        if debug:
            print("Channel=", channel)
        for block_type in ["datalogger", "preamplifier", "sensor"]:
            if block_type == "preamplifier" and block_type not in channel:
                print('component type "{}" absent, ignored'.format(block_type))
                continue  # to ignore empty  preamplifier
            if debug:
                print("Component=", block_type, channel[block_type])
            reference_code = channel[block_type]["reference_code"]
            serial_number = channel[block_type].get("serial_number", None)
            if debug:
                print("components=", components)

            channel[block_type] = components.get_component(
                block_type, reference_code, serial_number
            )
            if debug:
                print("channel[block_type]========",channel[block_type])
            if not channel[block_type]:
                raise NameError(
                    "Component not found: type:{}, "
                    "reference_code:{}, serial_number:{}".format(
                        block_type, reference_code, serial_number
                    )
                )

    def modify_sensors(self, sensor_dict, referring_file=None):
        """ Modify sensors within an instrument
        Inputs:
            sensor_dict: dictionary with key = component, val=[reference_code, serial_number]
        """

        components, path = load_information_file(
            self,
            components_file + root_symbol + "instrument_components/instrument_blocks",
            referring_file,
        )
        sensors_generic = components["sensor"]["generic"]
        sensors_specific = components["sensor"]["specific"]

        for channel, values in sensor_dict.items():
            ref_code = values["reference_code"]
            SN = values["serial_number"]
            self.channels[channel]["sensor"] = sensors_generic[ref_code].copy()
            self.channels[channel]["sensor"]["serial_number"] = SN
            print("Setting {} {} SN to {}".format(channel, ref_code, SN))
            if ref_code in sensors_specific:
                if SN in sensors_specific[ref_code]:
                    for key, value in sensors_specific[ref_code][SN].items():
                        print("Setting {} to {}".format(key, value))
                        self.channels[channel]["sensor"][key] = value

    def fill_responses(self, debug=False):
        for name, channel in self.das_components.items():
            if debug:
                print(yaml.dump(channel))
            channel["sensor"].fill_responses()
            if "preamplifier" in channel:
                channel["preamplifier"].fill_responses()
            channel["datalogger"].fill_responses()
            if debug:
                print(
                    "channel {} has : {:d} sensor,{:d} preamp,{:d} logger stages".format(
                        name,
                        len(channel["sensor"].response),
                        len(channel["preamplifier"].response),
                        len(channel["datalogger"].response),
                    )
                )
                
            channel["response"] = []
            channel["response"].append(channel["sensor"].response)
            if "preamplifier" in channel:
                channel["response"].append(channel["preamplifier"].response)
            channel["response"].append(channel["datalogger"].response)




import obspy.core.inventory.util as obspy_util


def create_comments(temp):

    # gerer le cas de str et dict (pour le champ processing)
    if type(temp) in [str, dict]:
        return [obspy_util.Comment(temp)]
    comments = []
    for id, comment in enumerate(temp):
        value = comment["value"] if "value" in comment else comment
        begin_effective_time = (
            comment["BeginEffectiveTime"] if "BeginEffectiveTime" in comment else None
        )
        end_effective_time = (
            comment["EndEffectiveTime"] if "EndEffectiveTime" in comment else None
        )
        authors = create_authors(comment) if "authors" in comment else None
        comments.append(
            obspy_util.Comment(
                value,
                begin_effective_time=begin_effective_time,
                end_effective_time=end_effective_time,
                authors=authors,
            )
        )
    return comments


def create_authors(comment):

    authors = []
    for author in comment["authors"]:
        names = list(author["names"]) if "names" in author else None
        emails = list(author["emails"]) if "emails" in author else None
        agencies = list(author["agencies"]) if "agencies" in author else None
        phones = create_phoneNumber(author["phones"]) if "phones" in author else None
        authors.append(obspy_util.Person(names, agencies, emails, phones))

    return authors


def create_phoneNumber(phones):
    phonesObjects = []
    for phone in phones:
        country_code = phone["countryCode"] if "countryCode" in phone else None
        area_code = phone["areaCode"] if "areaCode" in phone else None
        phone_number = phone["phoneNumber"] if "phoneNumber" in phone else None
        description = phone["description"] if "description" in phone else None
        phonesObjects.append(
            obspy_util.PhoneNumber(area_code, phone_number, country_code, description)
        )

    return phonesObjects



""" 
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
    
    version 0.99
    
I need to modify the code so that it treats a $ref as a placeholder for the associated object
"""
# Standard library modules
import math as m
import os.path
import sys

# Non-standard modules
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

################################################################################
# Miscellaneous Routines


def calc_norm_factor(zeros, poles, norm_freq, pz_type, debug=False):
    """
    Calculate the normalization factor for give poles-zeros
    
    The norm factor A0 is calculated such that
                       sequence_product_over_n(s - zero_n)
            A0 * abs(------------------------------------------) === 1
                       sequence_product_over_m(s - pole_m)

    for s_f=i*2pi*f if the transfer function is in radians
            i*f     if the transfer funtion is in Hertz
    """

    A0 = 1.0 + 1j * 0.0
    if pz_type == "LAPLACE (HERTZ)":
        s = 1j * norm_freq
    elif pz_type == "LAPLACE (RADIANS/SECOND)":
        s = 1j * 2 * m.pi * norm_freq
    else:
        print(
            "Don't know how to calculate normalization factor for z-transform poles and zeros!"
        )
    for p in poles:
        A0 = A0 * (s - p)
    for z in zeros:
        A0 = A0 / (s - z)

    if debug:
        print("poles=", poles, ", zeros=", zeros, "s={:g}, A0={:g}".format(s, A0))

    A0 = abs(A0)

    return A0


##################################################
def round_down_minute(date_time, min_offset):
    """
    Round down to nearest minute that is at least minimum_offset seconds earlier
    """
    dt = date_time - min_offset
    dt.second = 0
    dt.microsecond = 0
    return dt


##################################################
def round_up_minute(date_time, min_offset):
    """
    Round up to nearest minute that is at least minimum_offset seconds later
    """
    dt = date_time + 60 + min_offset
    dt.second = 0
    dt.microsecond = 0
    return dt


##################################################
def make_channel_code(
    channel_seed_codes,
    given_band_code,
    instrument_code,
    orientation_code,
    sample_rate,
    debug=False,
):
    """
        Make a channel code from sensor, instrument and network information
        
        channel_seed_codes is a dictionary from the instrument_component file
        given_band, instrument, and orientation codes are from the network file
        sample_rate is from the network_file
    """
    band_base = channel_seed_codes["band_base"]
    if not len(band_base) == 1:
        raise NameError("Band code is not a single letter: {}".format(band_code))
    if not instrument_code == channel_seed_codes["instrument"]:
        raise NameError(
            "instrument and component instrument_codes do not "
            "match: {}!={}".format(inst_code, channel_seed_codes["instrument"])
        )
    if not orientation_code in [key for key in channel_seed_codes["orientation"]]:
        raise NameError(
            "instrument and component orientation_codes do not "
            "match: {}!={}".format(orientation_code, channel_seed_codes["orientation"])
        )
    if band_base in "FCHBMLVURPTQ":
        if sample_rate >= 1000:
            band_code = "F"
        elif sample_rate >= 250:
            band_code = "C"
        elif sample_rate >= 80:
            band_code = "H"
        elif sample_rate >= 10:
            band_code = "B"
        elif sample_rate > 1:
            band_code = "M"
        elif sample_rate > 0.3:
            band_code = "L"
        elif sample_rate > 0.03:
            band_code = "V"
        elif sample_rate > 0.003:
            band_code = "U"
        elif sample_rate >= 0.0001:
            band_code = "R"
        elif sample_rate >= 0.00001:
            band_code = "P"
        elif sample_rate >= 0.000001:
            band_code = "T"
        else:
            band_code = "Q"
    elif band_base in "GDES":
        if sample_rate >= 1000:
            band_code = "G"
        elif sample_rate >= 250:
            band_code = "D"
        elif sample_rate >= 80:
            band_code = "E"
        elif sample_rate >= 10:
            band_code = "S"
        else:
            raise ValueError("Short period instrument has sample rate < 10 sps")
    else:
        raise NameError("Unknown band code: {}".format(band_code))
    if band_code != given_band_code:
        raise NameError(
            "Band code calculated from component and sample rate"
            " does not match that given in network file: {} versus {}".format(
                band_code, given_band_code
            )
        )
    if debug:
        print(band_code)
    channel_code = band_code + instrument_code + orientation_code
    return channel_code


##################################################
def get_azimuth_dip(channel_seed_codes, orientation_code):
    " Returns azimuth and dip [value,error] pairs "

    if orientation_code in channel_seed_codes["orientation"]:
        azimuth = channel_seed_codes["orientation"][orientation_code]["azimuth.deg"]
        azimuth = [float(x) for x in azimuth]
        dip = channel_seed_codes["orientation"][orientation_code]["dip.deg"]
        dip = [float(x) for x in dip]
    else:
        raise NameError(
            'orientation code "{}" not found in '
            "component seed_codes.orientation".format(orientation_code)
        )
    return azimuth, dip



""" 
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
    
    version 0.99
    
I need to modify the code so that it treats a $ref as a placeholder for the associated object
"""
# Standard library modules
import math as m
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

from ..misc.info_files import load_information_file, root_symbol
from ..misc.FDSN import equipment_type as FDSN_equipment_type

################################################################################
class instrument_component:
    """ One obsinfo instrument component 
    
        Inputs:
            component_dict: generic component dictionary from 
                            instrument_components information file
            
    """

    def __init__(
        self,
        component_dict,
        basepath,
        component_type=None,
        reference_code=None,
        sample_rate = None,
        debug=False,
    ):
        """ Inputs:
                component_dict: generic component dictionary from 
                                instrument_components information file
                basepath: full path of directory containing Instrument_Components file
                component_type = component type ('datalogger','preamplifier' or 'sensor')
                reference_code = component reference code
            
        """
        if debug:
            print(basepath)
        self.basepath = basepath
        self.equipment = FDSN_equipment_type(component_dict["equipment"])
        if "seed_codes" in component_dict:
            self.seed_codes = component_dict["seed_codes"]
        self.response_superstages = component_dict["response_stages"]
        self.type = component_type
        self.reference_code = reference_code
        self.sample_rate = sample_rate 
        self.response = None

    def __repr__(self):
        return "<OBS_Instrument_Component: {}>".format(self.reference_code)

    def fill_responses(self, debug=False):
        """ Fill in instrument responses from references"""
        if debug:
            print("self.response_superstages=", end="")
            print(yaml.dump(self.response_superstages))
        self.__read_response_yamls()
        if debug:
            print("self.response=", end="")
            print(yaml.dump(self.response))

    def __read_response_yamls(self, debug=False):
        """ READ INSTRUMENT RESPONSES FROM RESPONSE_YAML FILES
    
        Input:
            directory: base directory of response_yaml files
        """
        self.response = {'decimation_info':None,'stages':list()}
        if debug:
            print(
                "{:d} superstage files for the component".format(
                    len(self.response_superstages)
                )
            )
        for superstage in self.response_superstages:
            superstage_file = superstage["$ref"]
            if debug:
                print("Reading superstage file {}".format(superstage_file))
            response, temp = load_information_file(
                superstage_file + root_symbol + "response", self.basepath
            )
            self.response['decimation_info'] = response['decimation_info'] if 'decimation_info' in response else None
            for stage in response['stages']:
                # IF STAGE FILTER IS A "$ref", READ AND INJECT THE REFERRED FILE
                if "$ref" in stage["filter"]:
                    # READ REFERRED FILE
                    filter_ref = os.path.join(
                        os.path.split(superstage_file)[0], stage["filter"]["$ref"]
                    )
                    if debug:
                        print("filter file ref:", filter_ref)
                    filter, temp = load_information_file(filter_ref, self.basepath)
                    # MAKE SURE IT'S THE SAME TYPE, IF SO INJECT
                    stage["filter"] = filter
                self.response['stages'].append(stage)
            if debug:
                print("{:d} stages read".format(len(stages)))
        if debug:
            print("{:d} total stages in component".format(len(self.response)))
            print(yaml.dump(self.response))

    def __read_response_yamls_old(self, debug=False):
        """ READ INSTRUMENT RESPONSES FROM RESPONSE_YAML FILES
    
        Input:
            directory: base directory of response_yaml files
        """
        self.response = list()
        if debug:
            print(
                "{:d} superstage files for the component".format(
                    len(self.response_superstages)
                )
            )
        for superstage in self.response_superstages:
            superstage_file = superstage["$ref"]
            if debug:
                print("Reading superstage file {}".format(superstage_file))
            stages, temp = load_information_file(
                superstage_file + root_symbol + "response/stages", self.basepath
            )
            for stage in stages:
                # IF STAGE FILTER IS A "$ref", READ AND INJECT THE REFERRED FILE
                if "$ref" in stage["filter"]:
                    # READ REFERRED FILE
                    filter_ref = os.path.join(
                        os.path.split(superstage_file)[0], stage["filter"]["$ref"]
                    )
                    if debug:
                        print("filter file ref:", filter_ref)
                    filter, temp = load_information_file(filter_ref, self.basepath)
                    # MAKE SURE IT'S THE SAME TYPE, IF SO INJECT
                    stage["filter"] = filter
                self.response.append(stage)
            if debug:
                print("{:d} stages read".format(len(stages)))
        if debug:
            print("{:d} total stages in component".format(len(self.response)))
            print(yaml.dump(self.response))



"""
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments

I need to modify the code so that it treats a $ref as a placeholder for the
associated object
"""
# Standard library modules
import os.path

# Non-standard modules
import obspy.core.inventory as obspy_inventory
import yaml
# import obspy.core.inventory.util as obspy_util
# from obspy.core.utcdatetime import UTCDateTime

from ..misc.info_files import load_information_file
from ..misc import FDSN as oi_FDSN
from .station import station as oi_station
from .util import create_comments

###############################################################################


class network:
    """ Everything contained in a network.yaml file

        Has two subclasses:
            stations (.station)
            network_info (..misc.network_info)
    """

    def __init__(self, filename, referring_file=None, debug=False):
        """ Reads from a network information file

        should also be able to specify whether or not it has read its sub_file
        """
        root, path = load_information_file(filename, referring_file)
        self.basepath = path
        self.revision = root["revision"].copy()
        self.format_version = root["format_version"]
        net = root["network"]
        self.facility_ref_name = net["facility"]["reference_name"]
        self.facility_full_name = net["facility"].get("full_name", None)
        self.campaign = net["campaign_reference_name"]
        self.network_info = oi_FDSN.network_info(net["general_information"])
        self.instrumentation_file = net["instrumentation"]
        if not self.instrumentation_file["$ref"]:
            print(
                "No instrumentation file specfied, cannot create StationXML"
            )
        self.stations = dict()
        if debug:
            print("in network:__init__()")
        for code, station in root["network"]["stations"].items():
            if debug:
                print(f"net={self.network_info.code},station={code}")
            self.stations[code] = oi_station(station, code,
                                             self.network_info.code)
            if self.instrumentation_file["$ref"]:
                # Fill the instrument
                self.stations[code].fill_instrument(
                    self.instrumentation_file, referring_file=self.basepath
                )

            if debug:
                print(self.stations[code])

    def __repr__(self):
        return "<{}: code={}, facility={}, campaign={}, {:d} stations>".format(
            __name__,
            self.network_info.code,
            self.facility_ref_name,
            self.campaign,
            len(self.stations),
        )

    def __make_obspy_inventory(self, stations=None, source=None, debug=False):
        """
        Make an obspy inventory object with a subset of stations

        stations = list of obs-info.OBS-Station objects
        source  =  value to put in inventory.source
        """
        my_net = self.__make_obspy_network(stations)
        if not source:
            if self.facility_full_name:
                source = self.facility_full_name
            else:
                source = (
                    self.revision["author"]["first_name"]
                    + " "
                    + self.revision["author"]["last_name"]
                )
        my_inv = obspy_inventory.inventory.Inventory([my_net], source)
        return my_inv

    def __make_obspy_network(self, stations, debug=False):
        """Make an obspy network object with a subset of stations"""
        obspy_stations = []
        for station in stations:
            obspy_stations.append(station.make_obspy_station())

        temp = self.network_info.comments
        comments = None
        if temp:
            comments = create_comments(temp)
        my_net = obspy_inventory.network.Network(
            self.network_info.code,
            obspy_stations,
            description=self.network_info.description,
            comments=comments,
            start_date=self.network_info.start_date,
            end_date=self.network_info.end_date,
        )
        return my_net

    def write_stationXML(self, station_name, destination_folder=None,
                         debug=False):
        station = self.stations[station_name]
        if debug:
            print("Creating obsPy inventory object")
        my_inv = self.__make_obspy_inventory([station])
        if debug:
            print(yaml.dump(my_inv))
        if not destination_folder:
            destination_folder = "."
        fname = os.path.join(
            destination_folder,
            "{}.{}.STATION.xml".format(self.network_info.code, station_name),
        )
        print("Writing to", fname)
        my_inv.write(fname, "STATIONXML")

    def write_station_XMLs(self, destination_folder=None):
        for station_name in self.stations:
            self.write_station(station_name, destination_folder)


def _make_stationXML_script(argv=None):
    """
    Creates StationXML files from a network file and instrumentation file tree

    """
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="obsinfo-makeSTATIONXML", description=__doc__)
    parser.add_argument("network_file", help="Network information file")
    parser.add_argument(
        "-d", "--dest_path", help="Destination folder for StationXML files"
    )
    # parser.add_argument( '-v', '--verbose',action="store_true",
    #            help='increase output verbosiy')

    args = parser.parse_args(argv)

    if args.dest_path:
        if not os.path.exists(args.dest_path):
            os.mkdir(args.dest_path)

    # READ IN NETWORK INFORMATION
    net = network(args.network_file)
    print(net)

    for station in net.stations:
        net.write_stationXML(station, args.dest_path)



iris_units = [
    "meter",
    "m",
    "m/s",
    "m/s**2",
    "centimeter",
    "cm",
    "cm/s",
    "cm/s**2",
    "millimeter",
    "mm",
    "mm/s",
    "mm/s**2",
    "mm/hour",
    "micrometer",
    "um",
    "um/s",
    "um/s**2",
    "nanometer",
    "nm",
    "nm/s",
    "nm/s**2",
    "second",
    "s",
    "millisecond",
    "ms",
    "microsecond",
    "us",
    "nanosecond",
    "ns",
    "minute",
    "min",
    "hour",
    "radian",
    "rad",
    "microradian",
    "urad",
    "nanoradian",
    "nrad",
    "rad/s",
    "rad/s**2",
    "degree",
    "deg",
    "kelvin",
    "K",
    "celsius",
    "degC",
    "candela",
    "cd",
    "pascal",
    "Pa",
    "kilopascal",
    "kPa",
    "hectopascal",
    "hPa",
    "bar",
    "bars",
    "millibar",
    "mbar",
    "ampere",
    "amperes",
    "A",
    "milliamp",
    "mA",
    "volt",
    "V",
    "millivolt",
    "mV",
    "microvolt",
    "uV",
    "ohm",
    "hertz",
    "Hz",
    "newton",
    "N",
    "joule",
    "J",
    "tesla",
    "T",
    "nanotesla",
    "nT",
    "strain",
    "m/m",
    "m**3/m**3",
    "cm/cm",
    "mm/mm",
    "um/um",
    "nm/nm",
    "microstrain",
    "watt",
    "W",
    "milliwatt",
    "mW",
    "V/m",
    "W/m**2",
    "gap",
    "reboot",
    "byte",
    "bit",
    "bit/s",
    "percent",
    "%",
    "count",
    "counts",
    "number",
    "unitless",
    "unknown",
]



""" 
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
    
    version 0.99
    
I need to modify the code so that it treats a $ref as a placeholder for the associated object
"""
# Standard library modules
import math as m
import json
import pprint
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

from .misc import calc_norm_factor
from ..network.util import create_comments

################################################################################
# OBSPY-specific

last_output = None
def response_with_sensitivity(resp_stages, sensitivity, debug=False):

    true_sensitivity_input_units = None

    # HAVE TO MAKE OBSPY THINK ITS M/S FOR IT TO CALCULATE SENSITIVITY CORRECTLY FOR PRESSURE
    if "PA" in sensitivity["input_units"].upper():
        true_sensitivity_input_units = sensitivity["input_units"]
        sensitivity["input_units"] = "M/S"
    response = inventory.response.Response(
        instrument_sensitivity=inventory.response.InstrumentSensitivity(
            sensitivity["guess"],
            sensitivity["freq"],
            input_units=sensitivity["input_units"],
            output_units=sensitivity["output_units"],
            input_units_description=sensitivity["input_units_description"],
            output_units_description=sensitivity["output_units_description"],
        ),
        response_stages=resp_stages,
    )
    # response.plot(min_freq=0.001)
    guesstimate = response.instrument_sensitivity.value
    response.recalculate_overall_sensitivity(sensitivity["freq"])
    if debug:
        calculated = response.instrument_sensitivity.value
        print(
            "Guesstimated vs calculated sensitivity at {:g} Hz : {:.3g} vs {:.3g} ({:.1g}% difference)".format(
                response.instrument_sensitivity.frequency,
                guesstimate,
                calculated,
                100.0 * abs(guesstimate - calculated) / calculated,
            )
        )
    if true_sensitivity_input_units:
        response.instrument_sensitivity.input_units = true_sensitivity_input_units

    return response


def response(my_responses, debug=False):
    """
    Create an obspy response object from a response_yaml-based list of stages
    """
    global last_output
    resp_stages = []
    i_stage = 0
    sensitivity = dict()
    if debug:
        print(len(my_responses), "stages") #??
    #delay_correction = my_response.get("delay_correction",False)
    nbrS = get_nb_stages(my_responses)
    for my_response in my_responses:
        # temporary
        if my_response['decimation_info'] :
            key_with_p = [k for k in my_response['decimation_info'].keys() if '.' in k ]
            if len(key_with_p) !=0:
                delay_correction = my_response['decimation_info'][key_with_p[0]]
            else:
                delay_correction = my_response['decimation_info']['delay_correction'] if my_response['decimation_info'] else False
            if 'input_sample_rate' in my_response['decimation_info']:
                last_output = my_response['decimation_info']['input_sample_rate'] 
        decimation_info = my_response['decimation_info'] if 'decimation_info' in my_response else None

        for stage in my_response['stages']:
            # DEFINE COMMON VALUES
            i_stage = i_stage + 1
            #         if debug:
            #             print("stage=",end='')
            #             print(yaml.dump(stage))

            units, sensitivity = __get_units_sensitivity(stage, sensitivity, i_stage)

            resp_type = stage["filter"]["type"]
            if debug:
                print("i_stage=", i_stage, ", resp_type=", resp_type)
            # Create and append the appropriate response
            if resp_type == "PolesZeros":
                resp_stages.append(__make_poles_zeros(stage, i_stage, units))
            elif resp_type == "COEFFICIENTS":
                resp_stages.append(__make_coefficients(stage, i_stage, units, delay_correction, decimation_info, nbrS))
            elif resp_type == "FIR":
                resp_stages.append(__make_FIR(stage, i_stage, units, delay_correction, decimation_info, nbrS))
            elif resp_type == "AD_CONVERSION":
                resp_stages.append(__make_DIGITAL(stage, i_stage, units, decimation_info))
            elif resp_type == "ANALOG":
                resp_stages.append(__make_ANALOG(stage, i_stage, units))
            else:
                raise TypeError("UNKNOWN STAGE RESPONSE TYPE: {}".format(resp_type))
    response = response_with_sensitivity(resp_stages, sensitivity)
    if debug:
        print(response)
    return response


def get_nb_stages(responses):
    s = 0
    for i in responses:
        s = s + len(i['stages'])
    return s



def __get_units_sensitivity(stage, sensitivity, i_stage):
    # Get Units
    
    units = dict()
    temp = stage.get("input_units", {})
    units["input"] = temp.get("name", None)
    units["input_description"] = temp.get("description", None)

    temp = stage.get("output_units", {})
    units["output"] = temp.get("name", None)
    units["output_description"] = temp.get("description", None)

    # Set Sensitivity
    gain_value, gain_frequency = __get_gain(stage)
    if i_stage == 1:
        sensitivity = {
            "input_units": units["input"],
            "input_units_description": units["input_description"],
            "freq": gain_frequency,
            "guess": gain_value,
        }

    else:
        sensitivity["guess"] = sensitivity["guess"] * gain_value
    if units["output"]:
        sensitivity["output_units"] = units["output"]
        sensitivity["output_units_description"] = units["output_description"]

    return units, sensitivity


def __make_poles_zeros(stage, i_stage, units, debug=False):
    gain_value, gain_frequency = __get_gain(stage)
    resp = stage["filter"]
    lstr = resp["units"].lower()
    if "hertz" in lstr or "hz" in lstr:
        pz_type = "LAPLACE (HERTZ)"
    elif "z-transform" in lstr or "digital" in lstr:
        pz_type = "DIGITAL (Z-TRANSFORM)"
    elif "rad" in lstr:
        pz_type = "LAPLACE (RADIANS/SECOND)"
    else:
        raise ValueError('Unknown PoleZero response type: "{}"'.format(lstr))
    zeros = [
        obspy_types.ComplexWithUncertainties(
            float(t[0]) + 1j * float(t[1]), lower_uncertainty=0.0, upper_uncertainty=0.0
        )
        for t in resp["zeros"]
    ]
    poles = [
        obspy_types.ComplexWithUncertainties(
            float(t[0]) + 1j * float(t[1]), lower_uncertainty=0.0, upper_uncertainty=0.0
        )
        for t in resp["poles"]
    ]
    if gain_frequency == 0:
        norm_freq = 1.0
    else:
        norm_freq = gain_frequency
    norm_factor = resp.get(
        "normalization_factor", calc_norm_factor(zeros, poles, norm_freq, pz_type)
    )
    if debug:
        print(
            "  Z=",
            zeros,
            " P=",
            poles,
            " A0={:g} at {:g} Hz".format(norm_factor, norm_freq),
        )
    return inventory.response.PolesZerosResponseStage(
        i_stage,
        gain_value,
        gain_frequency,
        units["input"],
        units["output"],
        pz_transfer_function_type=pz_type,
        normalization_frequency=norm_freq,
        normalization_factor=norm_factor,
        zeros=zeros,
        poles=poles,
        input_units_description=units["input_description"],
        output_units_description=units["output_description"],
        description=stage["description"] if "description" in stage else None,
    )


def __make_coefficients(stage, i_stage, units, delay_correction, decimation_info,  nbr_stages, debug=False):
    resp = stage["filter"]
    gain_value, gain_frequency = __get_gain(stage)
    decim = {"delay": None, "factor": 1, "offset": None, "input_sr": None}
    correction = None
    if resp["type"].lower() == "hertz":
        cf_type = "ANALOG (HERTZ)"
    elif resp["type"].lower() == "digital":
        cf_type = "DIGITAL"
        decim = __get_decim_parms(stage)
        if delay_correction is True:
            correction = decim["delay"]
        elif type(delay_correction) is float and i_stage == nbr_stages:
            correction = delay_correction
        else:
            correction = 0.0
    else:
        cf_type = "ANALOG (RADIANS/S)"
    return inventory.response.CoefficientsTypeResponseStage(
        i_stage,
        gain_value,
        gain_frequency,
        units["input"],
        units["output"],
        cf_type,
        numerator=float(resp["numerator"]),
        denominator=float(resp["denominator"]),
        input_units_description=units["input_description"],
        output_units_description=units["output_description"],
        description=stage["description"],
        decimation_input_sample_rate=decim["input_sr"],
        decimation_factor=decim["factor"],
        decimation_offset=decim["offset"],
        decimation_delay=decim["delay"],
        decimation_correction=correction,
    )


def __make_FIR(stage, i_stage, units, delay_correction, decimation_info, nbr_stages, debug=False):
    resp = stage["filter"]
    if debug:
        print(resp)
    gain_value, gain_frequency = __get_gain(stage)
    decim = __get_decim_parms(stage, decimation_info)
    if delay_correction is True:
        correction = decim["delay"]
    elif type(delay_correction) is float and i_stage == nbr_stages:
        correction = delay_correction
    else:
        correction = 0.0
    return inventory.response.FIRResponseStage(
        i_stage,
        gain_value,
        gain_frequency,
        "counts",
        "counts",
        symmetry=resp["symmetry"].upper(),
        coefficients=[
            obspy_types.FloatWithUncertaintiesAndUnit(x) for x in resp["coefficients"]
        ],
        input_units_description="Digital Counts",
        output_units_description="Digital Counts",
        description=stage["description"] if "description" in stage else None,
        decimation_input_sample_rate=decim["input_sr"],
        decimation_factor=decim["factor"],
        decimation_offset=decim["offset"],
        decimation_delay=decim["delay"],
        decimation_correction=correction,
    )


def __make_DIGITAL(stage, i_stage, units, decimation_info, debug=False):
    gain_value, gain_frequency = __get_gain(stage)
    decim = __get_decim_parms(stage, decimation_info)
    return inventory.response.CoefficientsTypeResponseStage(
        i_stage,
        gain_value,
        gain_frequency,
        units["input"],
        units["output"],
        "DIGITAL",
        numerator=[
            obspy_types.FloatWithUncertaintiesAndUnit(
                1.0, lower_uncertainty=0.0, upper_uncertainty=0.0
            )
        ],
        denominator=[],
        input_units_description=units["input_description"],
        output_units_description=units["output_description"],
        description=stage["description"] if "description" in stage else None,
        decimation_input_sample_rate=decim["input_sr"],
        decimation_factor=decim["factor"],
        decimation_offset=decim["offset"],
        decimation_delay=decim["delay"],
        decimation_correction=float(stage.get("decimation_correction", 0.0)),
    )


def __make_ANALOG(stage, i_stage, units, debug=False):
    gain_value, gain_frequency = __get_gain(stage)
    # Force to PolesZeros without Poles or Zeros
    return inventory.response.PolesZerosResponseStage(
        i_stage,
        gain_value,
        gain_frequency,
        units["input"],
        units["output"],
        input_units_description=units["input_description"],
        output_units_description=units["output_description"],
        description=stage["description"] if "description" in stage else None,
        pz_transfer_function_type="LAPLACE (HERTZ)",
        normalization_frequency=0.0,
        normalization_factor=1.0,
        zeros=[],
        poles=[],
    )


#     return inventory.response.ResponseStage(\
#             i_stage,
#             gain_value, gain_frequency,
#             input_units, output_units,
#             input_units_description=input_units_description,
#             output_units_description=output_units_description,
#             description=stage['description']
#         )


def __get_gain(stage):
    gain = stage.get("gain", {})
    return float(gain.get("value", 1.0)), float(gain.get("frequency", 0.0))


def __get_decim_parms(stage, decimation_info):
    global last_output

    decim = dict()
    decim["factor"] = int(stage.get("decimation_factor", 1))

    if 'output_sample_rate' in stage:
        decim["input_sr"] = decim["factor"] * stage["output_sample_rate"]  
    elif  decimation_info and 'input_sample_rate' in decimation_info.keys():
        decim["input_sr"] = last_output
        last_output = decim["input_sr"] / decim["factor"]

    else:
        raise RuntimeError(f'Your stage {stage} does not have a "output_sample_rate" ')

    filter = stage["filter"]
    decim["offset"] = int(filter.get("delay.samples", 0))
    # cas sismob
    if "delay" in stage:
        decim["delay"] = stage["delay"]
    else:
        decim["delay"] = float(
            filter.get("delay", float(decim["offset"]) / float(decim["input_sr"]))
        )
    return decim


def equipment(equipment, resource_id=None, debug=False):
    """
    Create obspy EquipmentType from obs_info.equipment
    
    """
    if type(equipment) == dict:
        equipment = FDSN_EquipmentType(equipment)
    if not equipment.description:
        raise RuntimeError('Your equipment variable does not have a "description"')
    obspy_equipment = obspy_util.Equipment(
        type=equipment.type,
        description=equipment.description,
        manufacturer=equipment.manufacturer,
        vendor=equipment.vendor,
        model=equipment.model,
        serial_number=equipment.serial_number,
        installation_date=equipment.installation_date,
        removal_date=equipment.removal_date,
        calibration_dates=equipment.calibration_date,
        resource_id=resource_id,
    )
    if debug:
        print(equipment)
        print(obspy_equipment)
    return obspy_equipment


def comments(comments, processing, supplements, loc_code, location, debug=False):
    """
    Create obspy comments from station information
    
    Also stuffs fields that are otherwise not put into StationXML:
         "supplement" elements as JSON strings, 
          "location:location_methods" 
    """
    obspy_comments = []

    # add varible to check if clock_corrections is alredy commented
    clock_cor_commented = False
    if debug:
        print("supplements=", end="")
        print(supplements)

    if comments:
        obspy_comments += create_comments(comments)
    if supplements:  # ??
        for key, val in supplements.items():
            obspy_comments += create_comments(json.dumps({key: val}))
    if processing:
        for element in processing:
            if "clock_corrections" in element:
                clock_cor_commented = True
                for key, val in element["clock_corrections"].items():
                    obspy_comments += create_comments(
                        json.dumps({"clock_correction": {key: val}})
                    )
            else:
                obspy_comments += create_comments(element)

        if not clock_cor_commented:
            obspy_comments += create_comments(json.dumps({"clock_correction": None}))
    loc_comment = 'Using location "{}"'.format(loc_code)
    if "localisation_method" in location:
        loc_comment = loc_comment + ", localised using : {}".format(
            location["localisation_method"]
        )
        obspy_comments += create_comments(loc_comment)
    return obspy_comments


def lon_lats(location, debug=False):
    """ Calculate obspy util.Latitude and util.Longitude"""
    longitude = float(location["position"]["lon"])
    latitude = float(location["position"]["lat"])
    meters_per_degree_lat = 1852.0 * 60.0
    meters_per_degree_lon = 1852.0 * 60.0 * m.cos(latitude * m.pi / 180.0)
    lat_uncert = location["uncertainties.m"]["lat"] / meters_per_degree_lat
    lon_uncert = location["uncertainties.m"]["lon"] / meters_per_degree_lon
    # REDUCE UNCERTAINTIES TO 3 SIGNIFICANT FIGURES
    lat_uncert = float("{:.3g}".format(lat_uncert))
    lon_uncert = float("{:.3g}".format(lon_uncert))
    if debug:
        print(
            "{:.3f}+-{:.5f}, {:.3f}+-{:.5f}".format(
                longitude, lon_uncert, latitude, lat_uncert
            )
        )
    obspy_lat = obspy_util.Latitude(
        latitude, lower_uncertainty=lat_uncert, upper_uncertainty=lat_uncert
    )
    obspy_lon = obspy_util.Longitude(
        longitude, lower_uncertainty=lon_uncert, upper_uncertainty=lon_uncert
    )
    return obspy_lon, obspy_lat



""" 
nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
"""
# Standard library modules
import math as m
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

# obsinfo modules
from ..misc.info_files import load_information_file
from ..misc.FDSN import equipment_type as FDSN_equipment_type
from .instrument_component import instrument_component as oi_instrument_component


################################################################################
class instrument_components:
    """ Everything in a instrument_components information file"""

    def __init__(self, filename, referring_file=None, debug=False):
        """ Read instrument_components information file"""
        temp, basepath = load_information_file(filename, referring_file)
        if debug:
            print(basepath)
        self.basepath = basepath
        self.filename = filename
        self.revision = temp["revision"]
        self.format_version = temp["format_version"]
        # self.facility_reference_name=temp['facility_reference_name']
        self.instrument_blocks = temp["instrument_components"]["instrument_blocks"]

    def __repr__(self):
        return "<instrument_components: {}>".format(self.filename)

    def __load_specific_component(self, component, serial_number, debug=False):
        if not "specific" in self.instrument_blocks[component.type]:
            return component
        elif not serial_number in self.instrument_blocks[component.type]["specific"]:
            return component
        specific = self.instrument_blocks[component.type]["specific"][serial_number]
        if debug:
            print("[{}][{}]=".format(component.type, serial_number))
            print("specific=", specific)
        if "response" in specific:
            for key in specific["response"]:
                component["response"][key] = specific["response"][key]
        if "equipment" in specific:
            component.equipment.merge(FDSN_equipment_type(specific["equipment"]))
        return component

    def __check_response_files(
        self, files_dict, component_dict, resp_dir, print_names, debug=False
    ):
        if debug:
            print(files_dict)
            print(component_dict)
        if "generic" in component_dict:
            for component in component_dict["generic"].values():
                files_dict = self.__test_response_files(
                    files_dict, component, resp_dir, print_names
                )
        else:
            print('No generic objects for component type "{}"'.format(block_type))
        if "specific" in component_dict:
            for model in component_dict["specific"].values():
                for component in model.values():
                    files_dict = self.__test_response_files(
                        files_dict, component, resp_dir, print_names
                    )
        return files_dict

    def __test_response_files(
        self, files_dict, component, resp_dir, print_names, debug=False
    ):
        #         if debug:
        #             print('')
        #             print(component)
        if "response_stages" in component:
            for file_ref in component["response_stages"]:
                if debug:
                    print(file_ref)
                filename = file_ref["$ref"]
                if filename in files_dict:
                    files_dict[filename]["n_cites"] = (
                        files_dict[filename]["n_cites"] + 1
                    )
                else:
                    files_dict[filename] = dict(n_cites=1, exists=None)
                    if not os.path.isfile(os.path.join(resp_dir, filename)):
                        files_dict[filename]["exists"] = False
                    else:
                        files_dict[filename]["exists"] = True
        return files_dict

    def get_component(
        self, block_type, reference_code, serial_number=None, get_responses=True
    ):
        """ get one instrument_component
    
            inputs:
                block_type: 'datalogger','preamplifier' or 'sensor'
                reference_code:
                serial_number: optional
                get_responses: return with responses filled in [True]
            output:
                instrument_component: OBS_Instrument_Component object 
                                            OR 
                                      list of possible component names
        """
        if reference_code not in self.instrument_blocks[block_type]["generic"]:
            code_list = []
            for generic_code in self.instrument_blocks[block_type]["generic"]:
                if (reference_code + "_") in generic_code:
                    code_list.append(generic_code)
            if not code_list:
                code_list = None
            return code_list
        # LOAD GENERIC COMPONENT
        sample_rate = self.instrument_blocks[block_type]["generic"][reference_code]["sample_rate"] if "sample_rate" in self.instrument_blocks[block_type]["generic"][reference_code] else None
        component = oi_instrument_component(
            self.instrument_blocks[block_type]["generic"][reference_code], self.basepath, sample_rate=sample_rate
        )

        component.reference_code = reference_code
        component.type = block_type
        if serial_number:
            component.equipment.serial_number = serial_number
            # LOAD SPECIFIC COMPONENT, IF IT EXISTS
            component = self.__load_specific_component(component, serial_number)
        return component

    def print_elements(self, elem_type):
        """ prints one type of InstrumentComponent (descriptions and  serial numbers)
        type = ('sensor', "preamplifier' or 'datalogger')
        """
        for key, element in sorted(
            self.instrument_blocks[elem_type]["generic"].items()
        ):
            description = None
            if "equipment" in element:
                equipment = element["equipment"]
                if type(equipment) == dict:
                    equipment = FDSN_equipment_type(equipment)
                description = equipment.description
            if not description:
                if "description" in element:
                    description = element["description"]
                else:
                    description = "None provided"
            SNs = []
            if "specific" in self.instrument_blocks[elem_type]:
                if key in self.instrument_blocks[elem_type]["specific"]:
                    SNs = sorted(self.instrument_blocks[elem_type]["specific"][key])
            # output={'    model':key,'description':description,'specified_serial_numbers':SNs}
            # print(yaml.dump(output,indent=10,width=80))
            print(f'{key}: description="{description}", specified_SNs={SNs}')

    def verify_individuals(self):
        """ Verify that all "specific" instruments have a generic counterpart
        
            returns true if so, false + error message if not
        """
        checksOut = True
        for block_type in sorted(self.instrument_blocks.keys()):
            subcomponents = self.instrument_blocks[block_type]
            if "specific" in subcomponents:
                no_error_for_type = True
                for model in subcomponents["specific"].keys():
                    if model not in subcomponents["generic"]:
                        if no_error_for_type:
                            print("  {:>15}: ".format(block_type), end="")
                            no_error_for_type = False
                        checksOut = False
                        print(
                            15 * " "
                            + '"{}" is in "specific" but not in \
                                    "generic"'.format(
                                model
                            )
                        )
        return checksOut

    def verify_source_files(self, print_names=False, debug=False):
        """ Verify that all response files exist
        
            returns true if so, false + error message if not
            also returns number of files listed
        """
        print("Searching for response files")
        files_dict = dict()
        for block_type in self.instrument_blocks:
            files_dict = self.__check_response_files(
                files_dict,
                self.instrument_blocks[block_type],
                self.basepath,
                print_names,
            )
        if debug:
            print(files_dict)
        total_files = 0
        total_found = 0
        total_cites = 0
        for filename, value in sorted(files_dict.items()):
            total_files = total_files + 1
            n_cites = value["n_cites"]
            total_cites = total_cites + n_cites
            if value["exists"]:
                total_found = total_found + 1
                if print_names:
                    if n_cites == 1:
                        print("        FOUND ( 1 cite ): {}".format(filename))
                    else:
                        print(
                            "        FOUND ({:2d} cites): {}".format(n_cites, filename)
                        )
            else:
                if n_cites == 1:
                    print("    NOT FOUND ( 1 cite ): {}".format(filename))
                else:
                    print("    NOT FOUND ({:2d} cites): {}".format(n_cites, filename))
        return total_files, total_found, total_cites



""" 
Create OCA-JSON object from obs-info network
"""
import obsinfo
import json

################################################################################
class network:
    """OCA network description format
    
    Now that the instrument_components no longer name the response_stages but
    give a list instead, will have to deduce the OCA-compatible terms:
        - In Sensor: first item is sensor, any other are ana_filts
        - In Preamplifier: any items are ana_filts
        - In Datalogger: last item is digital_filter, second to last is digitizer,
            any prior are ana_filts
    """

    def __init__(self, obs_network, debug=False):
        """
        Create an OCA network object (contains stations without component details)
        """
        for station_code, station in obs_network.stations.items():
            if debug:
                print(station_code)
                print("A:", station.comments)
            station.partial_fill_instrument(
                obs_network.instrumentation_file, referring_file=obs_network.basepath
            )
            if debug:
                print("B:", station.comments)
            # self.stations[station_code]=station
        if debug:
            for key, station in obs_network.stations.items():
                print(station.comments)

        if debug:
            for key, station in obs_network.stations.items():
                print("B-Ca:", key, station.comments)
        self.dict = dict()
        self.__make_network_header(obs_network)
        self.__make_authors(obs_network)

        if debug:
            for key, station in obs_network.stations.items():
                print("B-Cb:", key, station.comments)
        self.dict["stations"] = []
        for key, station in obs_network.stations.items():
            if debug:
                print("C:", key, station.comments)
            oca_station = self.__make_station_header(station)
            # oca_station={"open_date":'duh',"close_date":'der'}
            oca_station = self.__make_devices_and_channels(
                oca_station, station.instrument.das_components
            )
            self.dict["stations"].append(oca_station)

    #     def dict(self,debug=False):
    #         """Make an oca JSON string """
    #         return self.dict

    def __repr__(self):
        return "<OCA_Network: fdsn_code={}>".format(self.dict["network"]["fdsn_code"])

    def __make_network_header(self, obs_network):
        netinfo = obs_network.network_info
        # Forces authorship of comments to network.yaml revision author
        for comment in netinfo.comments:
            comments = {"value": comment, "authors": ["auth1"]}
        self.dict["network"] = {
            "fdsn_code": netinfo.code,
            "temporary": True,
            "open_date": netinfo.start_date.isoformat(),
            "close_date": netinfo.end_date.isoformat(),
            "description": netinfo.description,
            "comments": comments,
        }

    def __make_authors(self, obs_network):
        """ Fills in author(s) information"""
        self.dict["comment_authors"] = dict()
        i = 0
        for author in obs_network.revision["authors"]:
            i = i + 1
            phones = list()
            if "phones" in author:
                for phone in author["phones"]:
                    phones.append({"phone_number": phone}),
            au_dict = {
                "firstname": author["first_name"],
                "lastname": author["last_name"],
                "emails": [author.get("email", "")],
                "phones": phones,
            }
            self.dict["comment_authors"]["auth" + str(i)] = au_dict

    def __make_station_header(self, obs_station, debug=False):
        """ Makes OCA STATION HEADER
        returns station dictionary
         """
        oca_station = dict()
        if debug:
            print("comments=", obs_station.comments)
            print("nClock_corrections=", len(obs_station.clock_corrections))
            print("nSupplements=", len(obs_station.supplements))
        oca_comments = [x for x in obs_station.comments]
        if obs_station.supplements:
            for key, value in obs_station.supplements.items():
                if debug:
                    print(key)
                oca_comments.append(json.dumps({key: value}).replace('"', "'"))
        if obs_station.clock_corrections:
            for key, value in obs_station.clock_corrections.items():
                if debug:
                    print(key)
                oca_comments.append(json.dumps({key: value}).replace('"', "'"))
        if debug:
            print("----")

        sta_position = obs_station.locations[obs_station.location_code]["position"]
        oca_station = {
            "IR_code": obs_station.code,
            "site": obs_station.site,
            "comments": oca_comments,
            "open_date": obs_station.start_date,
            "close_date": obs_station.end_date,
            "longitude": sta_position[0],
            "latitude": sta_position[1],
            "altitude": sta_position[2],
            "altitude_unit": "m",
            "first_install": obs_station.start_date,
        }
        return oca_station

    def __make_devices_and_channels(self, oca_station, das_components):

        oca_station["channels"] = list()
        devices = dict()
        for das_code, values in das_components.items():
            devices, device_codes = self.__add_devices(devices, values)
            ###
            ch = self.__make_channel(
                das_code,
                device_codes,
                values,
                oca_station["open_date"],
                oca_station["close_date"],
            )
            oca_station["channels"].append(ch)
        oca_station["installed_devices"] = devices
        return oca_station

    def __add_devices(self, devices, values, debug=False):
        """
        Find devices or add to dictionary of "installed devices"
        
        devices= existing dictionary of devices
        values = values corresponding to this das_component
        """
        codes = dict()
        if debug:
            print("===", values)

        sensor = self.__make_devicedict("sensor", values["sensor"].equipment)
        azimuth, dip = obsinfo.get_azimuth_dip(
            values["sensor"].seed_codes, values["orientation_code"]
        )
        sensor["azimuth_error"] = azimuth[1]
        sensor["dip_error"] = dip[1]
        sensor["local_depth"] = 0
        sensor["vault"] = "seafloor"
        sensor["config"] = "single"
        devices, codes["sensor"] = self.__find_append_device(devices, sensor, "sensor")

        anafilter = self.__make_devicedict(
            "anafilter", values["preamplifier"].equipment
        )
        devices, codes["anafilter"] = self.__find_append_device(
            devices, anafilter, "anafilter"
        )

        das = self.__make_devicedict("das", values["datalogger"].equipment)
        devices, codes["das"] = self.__find_append_device(devices, das, "das")

        return devices, codes

    def __make_devicedict(self, metatype, equipment):
        # if type(equipment) == dict:
        #    equipment=FDSN_EquipmentType(equipment)
        SN = equipment.serial_number
        devicedict = {
            "metatype": metatype,
            "manufacturer": equipment.manufacturer,
            "model": equipment.model,
            "serial_number": "unknown" if SN is None else SN,
        }
        return devicedict

    def __find_append_device(self, devices, device, device_name):
        """ 
        Find or append device to dictionary of devices 
    
        device codes are '{device_name}_N', where N is a counter for
        that type of device   
        """
        n_matching_devices = 0
        for device_code, value in devices.items():
            if device_name not in device_code:
                continue
            else:
                n_matching_devices = n_matching_devices + 1
                if device == value:
                    return devices, device_code
        device_code = device_name + "_" + str(n_matching_devices + 1)
        devices[device_code] = device
        return devices, device_code

    def __make_channel(
        self, das_code, device_codes, values, open_date, close_date, debug=False
    ):
        if debug:
            print("das_code=", das_code)
            print("device_codes=", device_codes)
            print("values=", values)
        das = das_code.split("_")
        das_component = das[0]
        if len(das) > 1:
            das_connector = das[1]
        else:
            das_connector = "1"
        azimuth, dip = obsinfo.get_azimuth_dip(
            values["sensor"].seed_codes, values["orientation_code"]
        )
        ana_filter_list = []
        for comp_type in ["sensor", "preamplifier", "datalogger"]:
            if comp_type in values:
                if debug:
                    print(f"values[{comp_type}]=", values[comp_type])
        #                 if "analog_filter" in values[comp_type].response_files:
        #                     ana_filter_list.append(
        #                         {"analog_filter":values[comp_type].response_files['analog_filter'],
        #                         "component": "1"})
        ch = {
            "location_code": values["location_code"],
            "seed_code": obsinfo.make_channel_code(
                values["sensor"].seed_codes,
                values["band_code"],
                values["inst_code"],
                values["orientation_code"],
                values["sample_rate"],
            ),
            "sensor": device_codes["sensor"],
            "sensor_component": values["orientation_code"],
            "analog_filter_list": ana_filter_list,
            "das": device_codes["das"],
            "das_connector": das_connector,
            "das_component": das_component,
            "das_digital_filter": values["datalogger"].reference_code,
            "data_format": "STEIM1",
            "polarity_reversal": dip == "-90.0",
            "open_date": open_date,
            "close_date": close_date,
            "channel_flags": "CG",
        }
        return ch



""" 
Validate information files

"""
# Standard library modules
import json
import pprint
import os.path
import sys
import pkg_resources

# Non-standard modules
import jsonschema
import jsonref
import yaml

root_symbol = "#"
VALID_FORMATS = ["JSON", "YAML"]
VALID_TYPES = [
    "campaign",
    "network",
    "instrumentation",
    "instrument_components",
    "response",
    "filter",
]
################################################################################


def list_valid_types():
    """
    Returns a list of valid information file types
    """
    return VALID_TYPES


################################################################################


def get_information_file_type(filename):
    """
    Determines the type of a file, assuming that the filename is "*.{TYPE}.{SOMETHING}
    """

    the_type = filename.split(".")[-2].split("/")[-1].lower()
    if the_type in VALID_TYPES:
        return the_type
    print(f"Unknown type: {the_type}")
    sys.exit(1)


################################################################################
def validate(filename, format=None, type=None, verbose=False, quiet=False):
    """
    Validates a YAML or JSON file against schema
    type: "network", "instrumentation","response", "instrument_components","filter"
    format: "JSON" or "YAML"
    
    if type and/or format are not provided, tries to figure them out from the
    filename, which should be "*{TYPE}.{FORMAT}
    """

    if quiet:
        verbose = False

    if not type:
        type = get_information_file_type(filename)

    #     with open(filename,'r') as f:
    #         if format=='YAML':
    #             instance=yaml.safe_load(f)
    #         else:
    #             instance=json.load(f)
    instance = read_json_yaml(filename, format=format)

    SCHEMA_FILE = pkg_resources.resource_filename(
        "obsinfo", f"data/schemas/{type}.schema.json"
    )
    base_path = os.path.dirname(SCHEMA_FILE)
    base_uri = f"file://{base_path}/"
    with open(SCHEMA_FILE, "r") as f:
        try:
            schema = jsonref.loads(f.read(), base_uri=base_uri, jsonschema=True)
        except json.decoder.JSONDecodeError as e:
            print(f"JSONDecodeError: Error loading JSON schema file: {SCHEMA_FILE}")
            print(str(e))
            return
        except:
            print(f"Error loading JSON schema file: {SCHEMA_FILE}")
            print(sys.exc_info()[1])
            return

    # Lazily report all errors in the instance
    # ASSUMES SCHEMA IS DRAFT-04 (I couldn't get it to work otherwise)
    try:
        if verbose:
            print(f"instance = {filename}")
        elif not quiet:
            print(f"instance = {filename} ... ", end="")

        if verbose:
            print(f"schema =   {os.path.basename(SCHEMA_FILE)}")
            print("\tTesting schema ...", end="")

        v = jsonschema.Draft4Validator(schema)

        if verbose:
            print("OK")
            print("\tTesting instance ...", end="")
        if not v.is_valid(instance):
            if quiet:
                # IF HAVE TO PRINT ERROR MESSAGE, PRINT INTRO TOO
                print(f"instance = {filename}")
            else:
                print("")
            for error in sorted(v.iter_errors(instance), key=str):
                print("\t\t", end="")
                for elem in error.path:
                    print(f"['{elem}']", end="")
                print(f": {error.message}")
            print("\tFAILED")
        else:
            if not quiet:
                print("OK")
    except jsonschema.ValidationError as e:
        if quiet:
            # IF HAVE TO PRINT ERROR MESSAGE, PRINT INTRO TOO
            print(f"instance = {filename}")
        else:
            print("")
        print("\t" + e.message)

    return


################################################################################


def get_information_file_format(filename):
    """
    Determines if the information file is in JSON or YAML format
    
    Assumes that the filename is "*.{FORMAT}
    """

    format = filename.split(".")[-1].upper()
    if format in VALID_FORMATS:
        return format
    print("Unknown format: {format}")
    sys.exit(1)


##################################################
def read_json_yaml(filename, format=None, debug=False):
    """ Reads a JSON or YAML file """
    if not format:
        format = get_information_file_format(filename)

    with open(filename, "r") as f:
        if format == "YAML":
            try:
                element = yaml.safe_load(f)
            except:
                print(f"Error loading YAML file: {filename}")
                print(sys.exc_info()[1])
                return
        else:
            try:
                element = json.load(filename)
            except JSONDecodeError as e:
                print(f"JSONDecodeError: Error loading JSON file: {filename}")
                print(str(e))
                return
            except:
                print(f"Error loading JSON file: {filename}")
                print(sys.exc_info()[1])
                return

    return element


##################################################
def load_information_file(
    reference, source_file=None, root_symbol=root_symbol, debug=False
):
    """
    Loads all (or part) of an information file
    
    input:
        reference (str): path to the element (filename &/or internal element path)
        source_file (str): full path of referring file (if any)
    output:
        element: the requested element
        base_file: the path of this file
        
    root_symbol is interpreted as the file's root level
     - If it is at the beginning of the reference, the element is searched for
        in source_file.
     - If it is in the middle of the reference, the element is searched for within the
        filename preceding it. 
     - If it is at the end (or absent), then the entire file is loaded 
     
    Based on JSON Pointers
       
    """

    # Figure out filename, absolute path and path inside file
    filename = None
    if root_symbol in reference:
        if reference.count(root_symbol) > 1:
            raise RuntimeError(
                'More than one occurence of "{}" in file reference "{}"'.format(
                    root_symbol, reference
                )
            )
        if reference[0] == root_symbol:
            filename = ""
            internal_path = reference[1:]
        elif reference[-1] == root_symbol:
            filename = reference[0:-1]
            internal_path = ""
        else:
            A = reference.split(root_symbol)
            filename = A[0]
            internal_path = A[1]
    else:
        filename = reference
        internal_path = ""
    if debug:
        print(
            "LOAD_INFORMATION_FILE(): reference={}, source_file={}".format(
                reference, source_file
            )
        )
    if source_file:
        if os.path.isfile(source_file):
            current_path = os.path.dirname(source_file)
        else:
            current_path = source_file
        filename = os.path.join(current_path, filename)
    else:
        current_path = os.getcwd()
    if debug:
        print(
            "LOAD_INFORMATION_FILE(): filename={}, internal_path={}".format(
                filename, internal_path
            )
        )

    # MAKE SURE THAT IT CONFORMS TO SCHEMA
    validate(filename, quiet=True)

    # READ IN FILE
    element = read_json_yaml(filename)

    # BREAK OUT THE REQUESTED PART
    if internal_path:
        for key in internal_path.split("/"):
            if not key in element:
                raise RuntimeError(
                    "Internal path {} not found in file {}".format(
                        internal_path, filename
                    )
                )
            else:
                element = element[key]

    # RETURN RESULT
    if debug:
        print("LOAD_YAML(): ", type(element))
    return element, os.path.abspath(os.path.dirname(filename))


################################################################################
def _validate_script(argv=None):
    """
    Validate an obsinfo information file

    Validates a file named *.{TYPE}.json or *.{TYPE}.yaml against the 
    obsinfo schema.{TYPE}.json file.

    {TYPE} can be campaign, network, instrumentation, instrument_components or
    response
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="obsinfo-validate", description=__doc__)
    parser.add_argument("info_file", help="Information file")
    parser.add_argument(
        "-t",
        "--type",
        choices=VALID_TYPES,
        default=None,
        help="Forces information file type (overrides interpreting from filename)",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=VALID_FORMATS,
        default=None,
        help="Forces information file format (overrides interpreting from filename)",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="increase output verbosiy"
    )
    args = parser.parse_args()

    validate(args.info_file, format=args.format, type=args.type, verbose=args.verbose)



""" 
Print complete stations from information in network.yaml file

nomenclature:
    A "measurement instrument" is a means of recording one physical parameter,
        from sensor through dac
    An "instrument" is composed of one or more measurement instruments
        
I need to modify the code so that it treats a $ref as a placeholder for the associated object
"""
# Standard library modules
import math as m
import json
import pprint
import os.path
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

# obsinfo modules
from ..misc.info_files import load_information_file, root_symbol
from ..misc import FDSN
from ..instrument_components import instrument_components as oi_instrument_components

################################################################################
class instrumentation:
    """ Everything in an instrumentation.yaml file
    
    Functions are very similar to those in instrument_components: should there 
    be a shared class?
    """

    def __init__(self, filename, referring_file=None):
        temp, basepath = load_information_file(filename, referring_file)

        self.basepath = basepath
        self.format_version = temp["format_version"]
        self.revision = temp["revision"]
        self.facility = temp["instrumentation"]["facility"]
        self.components_file = temp["instrumentation"]["instrument_components"]["$ref"]
        self.instruments = temp["instrumentation"]["instruments"]

    def __repr__(self):
        return "<OBS_Instrumentation: facility={}>".format(
            self.facility["reference_name"]
        )

    def __components_exist(
        self, components_dict, instrument, components, print_names, debug=False
    ):
        for das_component in instrument["das_components"].values():
            if debug:
                print(yaml.dump(das_component))
            for key, values in das_component.items():
                if "reference_code" in values:
                    ref_code = values["reference_code"]
                    if ref_code in components_dict:
                        components_dict[ref_code]["n_cites"] = (
                            components_dict[ref_code]["n_cites"] + 1
                        )
                    else:
                        components_dict[ref_code] = dict(
                            n_cites=1,
                            exists=None,
                            component_type=key,
                            config_list=False,
                        )
                        if debug:
                            print("components=", components)
                        component = components.get_component(key, ref_code)
                        if component:
                            components_dict[ref_code]["exists"] = True
                            if type(component) == list:
                                components_dict[ref_code]["config_list"] = [
                                    x[len(ref_code) + 1 :] for x in component
                                ]
                        else:
                            components_dict[ref_code]["exists"] = False
        return components_dict

    def print_elements(self, debug=False):
        """ prints all instruments (descriptions and  serial numbers)
        """
        for key, element in sorted(self.instruments["generic"].items()):
            description = None
            if debug:
                print(element)
            if "equipment" in element:
                equipment = element["equipment"]
                if type(equipment) == dict:
                    equipment = FDSN.equipment_type(equipment)
                description = equipment.description
            if not description:
                if "description" in element:
                    description = element["description"]
                else:
                    description = "None provided"
            SNs = []
            if "specific" in self.instruments:
                if key in self.instruments["specific"]:
                    SNs = sorted(self.instruments["specific"][key])
            # output={' model':key,'description':description,'specified_serial_numbers':SNs}
            # print(yaml.dump(output,width=76))
            print(f'{key}: description="{description}", specified_SNs={SNs}')

    def verify_individuals(self):
        """ Verify that all "specific" instruments have a generic counterpart
        
            returns true if so, false + error message if not
        """
        checksOut = True
        no_error_for_type = True
        if "specific" in self.instruments:
            for model in self.instruments["specific"].keys():
                if model not in self.instruments["generic"]:
                    if no_error_for_type:
                        print("  {:>15}: ".format(block_type), end="")
                        no_error_for_type = False
                    checksOut = False
                    print(
                        15 * " "
                        + '"{}" is in "specific" but not in "generic"'.format(model)
                    )
        return checksOut

    def check_dependencies(self, print_names=False, debug=False):
        """ Verify that the components file exists and contains requested components
        
            Prints out error message if anything fails
            returns:
                file_exists: true if file exists, false otherwise
                components_exist: true if all components exist, false otherwise
                n_components: number of components checked (including repeats)
        """
        comp_file = os.path.join(self.basepath, self.components_file)
        if not os.path.isfile(comp_file):
            print('Intrument_components file not found: "{}"'.format(comp_file))
            return False, False, None
        components = oi_instrument_components(comp_file)
        if debug:
            print("check_dependencies:components=", components)
        components_dict = dict()
        for instrument in self.instruments["generic"].values():
            if debug:
                print(yaml.dump(instrument))
            components_dict = self.__components_exist(
                components_dict, instrument, components, print_names
            )
        total_components = 0
        total_found = 0
        total_cites = 0
        for ref_code, value in sorted(components_dict.items()):
            total_components = total_components + 1
            n_cites = value["n_cites"]
            total_cites = total_cites + n_cites
            if value["exists"]:
                comments = ""
                total_found = total_found + 1
                if print_names:
                    if n_cites == 1:
                        print("        FOUND ( 1 cite ): {}".format(ref_code))
                    else:
                        print(
                            "        FOUND ({:2d} cites): {}".format(n_cites, ref_code)
                        )
            else:
                if n_cites == 1:
                    print("    NOT FOUND ( 1 cite ): {}".format(ref_code))
                else:
                    print("    NOT FOUND ({:2d} cites): {}".format(n_cites, ref_code))
        config_header_written = False
        for ref_code, value in sorted(components_dict.items()):
            if value["config_list"]:
                if not config_header_written:
                    print(
                        " **Some component citations must be completed in "
                        "the network.yaml file."
                    )
                    config_header_written = True
                print(
                    " **   - Stations using instruments containing: the {} {}".format(
                        ref_code, value["component_type"]
                    )
                )
                print(
                    " **                              must specify: {}_config".format(
                        value["component_type"]
                    )
                )
                print(
                    " **                   using one of the values: {}".format(
                        value["config_list"]
                    )
                )
        return True, total_components, total_found, total_cites



"""
station object in network information file
"""
# Standard library modules
import sys

# Non-standard modules
import yaml
import obspy.core.util.obspy_types as obspy_types
import obspy.core.inventory as obspy_inventory
import obspy.core.inventory.util as obspy_util
from obspy.core.utcdatetime import UTCDateTime

from ..misc import misc as oi_misc
from ..misc import obspy as oi_obspy
# from .util import create_comments
from ..misc.misc import make_channel_code
from ..instrumentation import instrument as oi_instrument


class station:
    """a station from the network information file"""

    def __init__(self, station_dict, station_code, network_code, debug=False):
        """ Create a station object directly from a network file's
        station: element """
        self.comments = station_dict.get("comments", [])
        self.site = station_dict["site"]
        self.start_date = station_dict["start_date"]
        self.end_date = station_dict["end_date"]
        self.instruments = station_dict["instruments"]
        self.station_location = station_dict["station_location"]
        self.locations = station_dict["locations"]
        self.processing = station_dict.get("processing", [])
        self.supplements = station_dict.get("supplements", [])  # ??
        self.code = station_code
        self.network_code = network_code
        if "sensors" in station_dict:
            self.sensors = station_dict["sensors"]
        else:
            self.sensors = None

    def __repr__(self):
        txt = "< {}: code={}, ".format(__name__, self.code)
        for inst in self.instruments:
            if hasattr(inst, "das_components"):
                txt += "instrument={} >".format(inst)
            else:
                txt += "instrument= ['{}','{}']".format(
                    inst["reference_code"], inst["serial_number"]
                )
        return txt

    def fill_instrument(self, instrument_file, referring_file=None):
        """ Fills in instrument information """
        self.partial_fill_instruments(instrument_file, referring_file)
        if self.sensors:
            print("Adding custom sensors")
            self.instruments.modify_sensors(self.sensors, referring_file)
        for inst in self.instruments:
            inst.fill_responses()

    def partial_fill_instruments(self, instrument_file,
                                 referring_file=None, debug=True):
        """ Converts network file instrument objects to Instrument class.
            ??? Does not fill in component responses ??? """
        instruments = []
        for inst_dict in self.instruments:
            inst = oi_instrument(instrument_file["$ref"], inst_dict,
                                 referring_file=referring_file)
            inst.load_components(inst.components_file, inst.basepath)
            self.operator = inst.facility  # à verifier??
            # instruments[inst_dict["reference_code"]]=inst
            instruments.append(inst)
        self.instruments = instruments

    def make_obspy_station(self, debug=False):
        """
        Create an obspy station object from a fully informed station
        """
        # CREATE CHANNELS

        # if debug:
        #    print(self)
        channels = []
        for instrument in self.instruments:
            # resource_id = instrument.resource_id
            for key, chan in instrument.das_components.items():
                if debug:
                    print(key)
                    print(yaml.dump(chan))
                response = oi_obspy.response(chan["response"])
                # loc_code=key.split(':')[1]
                loc_code = chan["location_code"]
                try:
                    location = self.locations[loc_code]
                except KeyError:
                    print(f"location code {loc_code} not found in ")
                    print("self.locations, valid keys are:")
                    for key in self.locations.keys():
                        print(key)
                    sys.exit(2)
                obspy_lon, obspy_lat = oi_obspy.lon_lats(location)
                azi, dip = oi_misc.get_azimuth_dip(
                    chan["sensor"].seed_codes, chan["orientation_code"]
                )
                start_date = None
                end_date = None
                start_date_chan = None
                end_date_chan = None
                # Give at least 3 seconds margin around start and end dates

                if "start_date" in chan:
                    start_date_chan = UTCDateTime(chan["start_date"])

                if self.start_date:
                    start_date = UTCDateTime(self.start_date)

                if "end_date" in chan:
                    end_date_chan = UTCDateTime(chan["end_date"])

                if self.end_date:

                    end_date = UTCDateTime(self.end_date)

                if debug:
                    print(key)
                    print(yaml.dump(chan))
                # print(location)
                if "localisation_method" in location:
                    channel_comment = obspy_util.Comment(
                        "Localised using : {}".format(
                                location["localisation_method"])
                    )
                else:
                    channel_comment = None
                channel_code = make_channel_code(
                    chan["sensor"].seed_codes,
                    chan["band_code"],
                    chan["inst_code"],
                    chan["orientation_code"],
                    chan["datalogger"].sample_rate,
                )
                start_date = start_date_chan if start_date_chan else start_date
                channel = obspy_inventory.channel.Channel(
                    code=channel_code,
                    location_code=loc_code,
                    latitude=obspy_lat,
                    longitude=obspy_lon,
                    elevation=obspy_types.FloatWithUncertaintiesAndUnit(
                        location["position"]["elev"],
                        lower_uncertainty=location["uncertainties.m"]["elev"],
                        upper_uncertainty=location["uncertainties.m"]["elev"],
                    ),
                    depth=location["depth.m"],
                    azimuth=obspy_types.FloatWithUncertainties(
                        azi[0],
                        lower_uncertainty=azi[1] if len(azi) == 2 else 0,
                        upper_uncertainty=azi[1] if len(azi) == 2 else 0,
                    ),
                    dip=dip[0],
                    types=["CONTINUOUS", "GEOPHYSICAL"],
                    sample_rate=chan["datalogger"].sample_rate,
                    clock_drift_in_seconds_per_sample=1
                    / (1e8 * float(chan["datalogger"].sample_rate)),
                    sensor=oi_obspy.equipment(chan["sensor"].equipment),
                    pre_amplifier=oi_obspy.equipment(
                                    chan["preamplifier"].equipment)
                    if "preamplifier" in chan
                    else None,
                    data_logger=oi_obspy.equipment(
                                    chan["datalogger"].equipment),
                    equipment=None,
                    response=response,
                    description=None,
                    comments=[channel_comment] if channel_comment else None,
                    start_date=start_date,
                    end_date=end_date_chan if end_date_chan else end_date,
                    restricted_status=None,
                    alternate_code=None,
                    data_availability=None,
                )
                channels.append(channel)
                if debug:
                    print(yaml.dump(channel))
            # CREATE STATION
            station_loc_code = self.station_location  # david
            if station_loc_code in self.locations:
                sta_loc = self.locations[station_loc_code]
                obspy_lon, obspy_lat = oi_obspy.lon_lats(sta_loc)
            else:
                print("No valid location code for station, either ", end='')
                print("set station_location_code or provide a location '00'")
                sys.exit()

            obspy_comments = oi_obspy.comments(
                self.comments,
                self.processing,
                self.supplements,
                station_loc_code,
                sta_loc,
            )

            # DEFINE Operator
            agency = self.operator["full_name"]
            contacts = None
            if "email" in self.operator:
                contacts = [obspy_util.Person(emails=[self.operator["email"]])]
            website = self.operator.get("website", None)
            operator = obspy_util.Operator([agency], contacts, website)

            if debug:
                print(obspy_comments)
            sta = obspy_inventory.station.Station(
                code=self.code,
                latitude=obspy_lat,
                longitude=obspy_lon,
                elevation=obspy_types.FloatWithUncertaintiesAndUnit(
                    sta_loc["position"]["elev"],
                    lower_uncertainty=sta_loc["uncertainties.m"]["elev"],
                    upper_uncertainty=sta_loc["uncertainties.m"]["elev"],
                ),
                channels=channels,
                site=obspy_util.Site(getattr(self, "site", "")),
                vault=sta_loc["vault"],
                geology=sta_loc["geology"],
                equipments=[
                    oi_obspy.equipment(instrument.equipment)
                    for instrument in self.instruments
                ],
                operators=[operator],
                creation_date=start_date,  # Needed to write StationXML
                termination_date=end_date,
                description=None,
                comments=obspy_comments,
                start_date=start_date if start_date else None,
                end_date=end_date if end_date else None,
                restricted_status=None,
                alternate_code=None,
                data_availability=None,
            )
        if debug:
            print(sta)
        return sta



""" 
Validate information files

"""
# Standard library modules
import pprint
import os.path
import sys

# obsinfo modules
from .info_files import (
    read_json_yaml,
    get_information_file_format,
    get_information_file_type,
    VALID_TYPES,
    VALID_FORMATS,
)
from ..network import network
from ..instrumentation import instrumentation
from ..instrument_components import instrument_components

################################################################################
def print_summary(filename, format=None, type=None, verbose=False, debug=False):
    """
    Print a summary of an information file
    type: "network", "instrumentation","response", "instrument_components","filter"
    format: "JSON" or "YAML"
    
    if type and/or format are not provided, figures them out from the
    filename, which should be "*{TYPE}.{FORMAT}
    """

    if not type:
        type = get_information_file_type(filename)

    instance = read_json_yaml(filename)

    print(f"\nFILENAME: {filename}")
    if type == "network":
        if debug:
            print("Loading network")
        instance = network(filename)
        if debug:
            print("Done")
        _print_summary_network(instance, filename)
    elif type == "instrumentation":
        instance = instrumentation(filename)
        _print_summary_instrumentation(instance)
    elif type == "instrument_components":
        instance = instrument_components(filename)
        _print_summary_instrument_components(instance)
    else:
        _print_summary_other(instance[type])


################################################################################
def _print_summary_network(network, network_file):
    """ Print summary information about a network """
    print(network)
    for name, station in network.stations.items():
        print("  " + str(station))
    print("")


################################################################################
def _print_summary_instrumentation(instrumentation):
    """ Print summary information about an instrumentation file """
    print(f"FACILITY: {instrumentation.facility['reference_name']}")
    print(f"REVISION: {instrumentation.revision}")
    # PRINT INSTRUMENTS
    print(20 * "=")
    print("INSTRUMENTS:")
    instrumentation.print_elements()

    # VERIFY THAT INSTRUMENTS & SENSORS LISTED IN "individuals" EXIST in "models"
    print(20 * "=")
    if instrumentation.verify_individuals():
        print("All instruments have a generic counterpart")

    # VERIFY THAT REFERRED TO FILES EXIST
    print(20 * "=")
    print("Checking dependencies on instrument_components_file")
    file_exists, n_components, n_found, n_cites = instrumentation.check_dependencies(
        print_names=True
    )
    if not file_exists:
        print(
            f"Instrument_Components file not found: {instrumentation.components_file}"
        )
    elif n_components == n_found:
        print(
            "Found all {:d} specied functional components ({:d} total cites)"
            "".format(n_components, n_cites)
        )
    else:
        print(
            "MISSING {:d} of {:d} specied functional components "
            "".format(n_components - n_found, n_components)
        )


################################################################################
def _print_summary_instrument_components(instance):
    """ Print summary information about an instrument_components file """
    print(f"FORMAT_VERSION: {instance.format_version}")
    print(f"REVISION: {instance.revision}")
    # PRINT INSTRUMENTS
    print(10 * "=")
    print("DATALOGGERS:")
    instance.print_elements("datalogger")
    print(10 * "=")
    print("PREAMPLIFIERS:")
    instance.print_elements("preamplifier")
    print(10 * "=")
    print("SENSORS:")
    instance.print_elements("sensor")

    # VERIFY THAT COMPONENTS LISTED IN "specific" EXIST in "generic"
    print(10 * "=")
    if instance.verify_individuals():
        print("All specific components have a generic counterpart")

    # VERIFY THAT REFERRED TO FILES EXIST
    print(10 * "=")
    n_files, n_found, n_cites = instance.verify_source_files(print_names=True)
    if n_files == n_found:
        print(
            "Found all {:d} specified source files ({:d} total citations)"
            "".format(n_files, n_cites)
        )
    else:
        print(
            "MISSING {:d} of {:d} specified source files"
            "".format(n_files - n_found, n_files)
        )


################################################################################
def _print_summary_other(instance):
    """ Print summary information about a generic information file """
    pp = pprint.PrettyPrinter(indent=1, depth=2, compact=True)
    pp.pprint(instance)


################################################################################
def _print_script(argv=None):
    """
    Prints summary of an information file

    Also confirms existence of referenced file(s)
    """
    from argparse import ArgumentParser

    parser = ArgumentParser(prog="obsinfo-print", description=__doc__)
    parser.add_argument("info_file", help="Information file")
    parser.add_argument(
        "-t",
        "--type",
        choices=VALID_TYPES,
        default=None,
        help="Forces information file type (overrides interpreting from filename)",
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=VALID_FORMATS,
        default=None,
        help="Forces information file format (overrides interpreting from filename)",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="increase output verbosiy"
    )
    args = parser.parse_args()

    print_summary(
        args.info_file, format=args.format, type=args.type, verbose=args.verbose
    )



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()
    
version={}
with open("obsinfo/version.py") as fp:
    exec(fp.read(),version)

setuptools.setup(
    name="obsinfo",
    version=version['__version__'],
    author="Wayne Crawford",
    author_email="crawford@ipgp.fr",
    description="Tools for documenting ocean bottom seismometer experiments and creating meta/data",
    long_description=long_description,
    long_description_content_type="text/x-rst; charset=UTF-8",
    url="https://github.com/WayneCrawford/obsinfo",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=[
          'obspy>=1.1',
          'pyyaml>=3.0',
          'jsonschema>=2.6',
          'jsonref>=0.2'
      ],
#     scripts=[
#         'scripts/obsinfo-print.py',
#         'scripts/obsinfo-makeSTATIONXML.py'
#     ],
    entry_points={
        'console_scripts': [
            'obsinfo-validate=obsinfo.misc.info_files:_validate_script',
            'obsinfo-print=obsinfo.misc.print:_print_script',
            'obsinfo-makeSTATIONXML=obsinfo.network.network:_make_stationXML_script',
            'obsinfo-make_SDPCHAIN_scripts=obsinfo.addons.SDPCHAIN:_console_script',
            'obsinfo-make_LCHEAPO_scripts=obsinfo.addons.LCHEAPO:_console_script'
        ]
    },
    python_requires='>=3',
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics"
    ),
    keywords='seismology OBS'
)
